import java.util.TreeMap;
import java.util.Iterator;
import java.util.Map;

/**
   Tom Austin, Arnold Suvatne
   SJSU - cs249
   
   LogicalClockBroadcastProcess.java
   Nov 25, 2005
*/

/**
   This is a broadcast process that uses logocal clocks for a partial
   ordering of events on the system.
   
   This is an extension of the broadcast algorithms by Vassos Hadzilacos
   and Sam Toueg presented in Chapter 5 of "Distributed Systems",
   2nd edition, 1993, Mullender.
   
   This does NOT use one of their algorithms, but uses their technique
   of layering new algorithms over the old.
   
   @author Tom Austin
 */
public class LogicalClockBroadcastProcess extends ReliableBroadcastProcess
{
   //This is used as the sequence # for outgoing messages instead of the
   // messagesReceived variable from ReliableBroadcastProcess.
   protected int internalLogicalClock;
   
   //These messages are considered delivered. 
   private Map<Integer,String> deliveredMessagesMap;
   
   /**
      Constructor.
    */
   public LogicalClockBroadcastProcess(String ID)
   {
      this(ID,0);
   }
   
   /**
      Constructor that specified the starting value for the logical clock.
    */
   public LogicalClockBroadcastProcess(String ID, int startingLogicalClockVal)
   {
      super(ID);
      internalLogicalClock = startingLogicalClockVal;
      deliveredMessagesMap = new TreeMap<Integer,String>();
   }
   
   /**
      As part of this algorithm, local steps are supposed to increment
      the logical time.  If a process is using this to handle its
      communication with other processes, this may be called to increment
      the clock time during local operations.
    */
   public void localStep()
   {
      internalLogicalClock++;
   }
   
   /**
      This increments the logical clock time and then broadcasts
      a message.
      
      @see ReliableBroadcastProcess#broadcast(java.lang.String)
    */
   public void broadcast(String messageText)
   {
      internalLogicalClock++;
      
      Message m = new Message(messageText, super.procID, internalLogicalClock);
      sendToAll(m);
   }
   
   /**
      This delivers a message to the system and updates
      the logical clock time if needed.  Unlike with
      the other Broadcast algorithms, nothing will be
      printed to standard out, unless the text of the
      message is "SCHED".
      
      @see ReliableBroadcastProcess#deliver(Message)
    */
   public void deliver(Message m)
   {
      //Local time is raised to the time on the message if that time
      // is in the future.
      int messageSeqNum = m.getSequenceNumber();
      if (internalLogicalClock < messageSeqNum)
         internalLogicalClock = messageSeqNum;
      
      //Time is incremented one either way.
      internalLogicalClock++;
      
      deliveredMessagesMap.put(internalLogicalClock, m.toString());
      
      //Messages of "SCHED" are taken as requests to
      // print out the current schedule.
      if (m.toString().equals("SCHED"))
         System.out.println(this.toString());
   }
   
   /**
      Returns a print out of the current schedule.
      
      @see java.lang.Object#toString()
    */
   public String toString()
   {
      String s = "\n***" + super.procID + " Event Ordering ***\n";
      
      Iterator<Integer> iter = deliveredMessagesMap.keySet().iterator();
      while (iter.hasNext())
      {
         int eventTime = iter.next();
         String message = deliveredMessagesMap.get(eventTime);
         
         s += "  " + eventTime + ": " + message + "\n";
      }
      
      return s;
   }
   
   
   /**
      Calls the start method for this class.
    */
   public static void main(String[] args)
   {
      if (args.length < 1)
      {
         System.out.println("Usage: java LogicalClockBroadcastProcess <process name>");
         System.exit(0);
      }
      
      System.out.println("Enter 'SCHED' to print out the current schedule.");
      LogicalClockBroadcastProcess lcbp = new LogicalClockBroadcastProcess(args[0]);
      lcbp.start();
   }
}
