import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

/**
  Tom Austin, Arnold Suvatne
  SJSU - cs249

  MutexProcess.java
  Nov 25, 2005
 */

/**
   This class simulates a process that may try to use a
   single shared resource.
   
   @author Tom Austin
 */
public class MutexProcess extends LogicalClockBroadcastProcess
{
   private static final int MAX_DELAY = 3000;
   
   private Random randomizer;
   
   //List of other processes's IDs.
   private List<String> processList;
   
   //Time that this process last sent a request to use the resource.
   private int lastRequestTime;
   
   //Queue of requests.
   private List<String> requestQueue;
   
   //List of IDs of the processes that have acknowledged the request.
   private List<String> ackList;
   
   
   
   /**
      This class simulates using a resource and handles the needed
      details of coordinating with the other processes.
      
      @author Tom Austin
    */
   private class ResourceUser extends Thread
   {
      /**
         This starts a request for a resource.
         
         @see java.lang.Runnable#run()
       */
      public void run()
      {
         getResource();
      }
      
      /**
         Requests a resource, "uses" it (by doing nothing and just holding
         on to it) and then releases the resource.
         
         @param resourceName Name of the resource requested.
       */
      private void getResource()
      {
         //Clear the acknowledgment list before broadcasting the request.
         ackList = new ArrayList<String>();
         broadcast("REQUEST");
         
         //Wait to be at the head of the queue
         boolean firstInLine = false;
         while (!firstInLine)
         {
            System.out.println(requestQueue);
            if (!requestQueue.isEmpty() && requestQueue.get(0).equals(procID))
            {
               firstInLine = true;
            }
            if (!firstInLine)
            {
               try
               {
                  Thread.sleep(1000);
               }
               catch (InterruptedException e)
               {
                  e.printStackTrace();
               }
            }
         }
         
         //Wait for the other processes to acknowledge.
         boolean allClearToProceed = false;
         while (!allClearToProceed)
         {
            //We set this to true.  If any process has not sent its OK,
            // then this will be set back to false.
            allClearToProceed = true;
            for (Iterator<String> iter = processList.iterator(); iter.hasNext(); )
            {
               String pID = iter.next();
               if (!ackList.contains(pID))
                  allClearToProceed = false;
            }
            
            //We sleep for a second if we have not received OKs from the other processes.
            if (!allClearToProceed)
            {
               try
               {
                  Thread.sleep(1000);
               }
               catch (InterruptedException e)
               {
                  e.printStackTrace();
               }
            }
         }
         
         //At this point, we have the resource, and so we use it for some random
         // period of time.
         try
         {
            String message = "***" + procID + " USING RESOURCE***";
            System.out.println(message);
            broadcast(message);
            int sleepTime = Math.abs(randomizer.nextInt()) % MAX_DELAY;
            Thread.sleep(sleepTime);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
         
         //Now that this process is finished using the resource, it removes itself
         // from the head of the queue and broadcasts a release to the other processes.
         requestQueue.remove(procID);
         broadcast("RELEASE");
      }
   }
   //End of ResourceUser class.
   
   
   
   /**
      Constructor.  Creates a new instance with a starting logical clock
      value of 0.  The name is the unique ID of this process.
    */
   public MutexProcess(String name)
   {
      this(name, 0);
   }
   
   /**
      Constructor.  Creates a new instance with the specified name
      and starting value for the logical clock.
    */
   public MutexProcess(String name, int logicalClockStartValue)
   {
      super(name, logicalClockStartValue);
      lastRequestTime = 0;
      randomizer = new Random();
      requestQueue = new ArrayList<String>();
      ackList = new ArrayList<String>();
      processList = new ArrayList<String>();
      processList.add(name);
   }
   
   /**
      This is identical to the broadcast of
      LogicalClockBroadcastProcess, except that it will keep
      track of the time when requests are made.
      
      @see ReliableBroadcastProcess#broadcast(java.lang.String)
    */
   public void broadcast(String messageText)
   {
      if (messageText.equals("REQUEST"))
         lastRequestTime = super.internalLogicalClock + 1;
      super.broadcast(messageText);
   }
   
   /**
      This will handle process requests and acknowledgements.
      @see LogicalClockBroadcastProcess#deliver(Message)
    */
   public void deliver(Message m)
   {
      super.deliver(m);
      
      String messageText = m.toString();
      String senderID = m.getSenderID();
      
      //Whenever we receive a message, we add the sender's name
      // to the acknowledgement list if the timestamp on the
      // message is later than the time of the last request.
      if (m.getSequenceNumber() > lastRequestTime)
         ackList.add(senderID);
      
      //Deal with requests.
      if (messageText.equals("REQUEST"))
      {
         System.out.println("Received request from " + senderID);
         requestQueue.add(senderID);
         broadcast("ACKNOWLEDGE");
      }
      
      //Deal with releases.
      if (messageText.equals("RELEASE"))
      {
         System.out.println("Received release from " + senderID);
         requestQueue.remove(senderID);
      }
      
      //Request from another process to join this network.
      if (messageText.equals("JOIN") && !processList.contains(senderID))
      {
         processList.add(senderID);
         //We broadcast a join request back so that the new process will
         // add this process to its list.
         broadcast("JOIN");
         System.out.println("Adding " + senderID + " to " + super.procID + "'s list.");
      }
      
      //When "RR" has been broadcast by this process, it triggers a
      // request for a resource.
      if (messageText.equals("RR") && senderID.equals(super.procID))
      {
         requestResource();
      }
   }
   
   public void requestResource()
   {
      ResourceUser resUser = new ResourceUser();
      resUser.start();
      System.out.println("\tStarted resource request.");
   }
   
   /**
      @see ReliableBroadcastProcess#startListenerThread()
    */
   protected void startListenerThread()
   {
      super.startListenerThread();
      broadcast("JOIN");
   }

   /**
      Calls the start method for this class.
    */
   public static void main(String[] args)
   {
      if (args.length < 1)
      {
         System.out.println("Usage: java MutexProcess <process name>");
         System.exit(0);
      }
      
      System.out.println("Enter RR to simulate a resource request and SCHED to print out the current schedule.");
      
      MutexProcess mp = new MutexProcess(args[0]);
      mp.start();
   }

}
